<#
.SYNOPSIS
    Synchronous client side socket interface for Temenos Connector Server.

.DESCRIPTION
    Takes a Temenos Connector request in the ofs/ofsml format and send it via TCP/IP
    using a socket interface. Response messages from 'ENQUIRY' request types have
    numerous data format transformation options like: csv, excel, html, pdf, etc,.
    Response messages from all other request types supported will be output as
    received from TCServer.

    Temenos Connector request types supported:
    - T24 APPLICATION (TRANSACTION)
    - ENQUIRY.SELECT
    - XML.REPORT
    - TEC.OFS.INTERFACE
    - CLEARING

.PARAMETER Ofstx
    Send request message in OFS format syntax.

.PARAMETER request
    Specifies the OFS request message to be sent.

.PARAMETER Ofsml
    Send request message in OFSML (xml) format syntax.

.PARAMETER path
    Specifies the path to the xml file that contains the OFSML request message to be sent.

.PARAMETER as
    Specifies the format the response message will be converted to.
    For OFS request types the options are: raw, plain, csv, excel and pdf.
    For OFSML request types the options are: raw and html.

.PARAMETER pretty
    Format XML response from an OFSML request into a human-readable form.

.PARAMETER channel
    Specifies 'CHANNEL' name in Channels.xml file where are found the settings that will
    be used for connecting to a local/remote TCServer listening on a BytesListener port.

.PARAMETER user
    Sign-on-name of the user that will be used to create the Temenos T24 session
    where the request message will be executed.

.PARAMETER password
    Password of the user that will be used to create the Temenos T24 session
    where the request message will be executed.

.PARAMETER company
    Company in the Temenos T24 session where the request message will be executed.

.PARAMETER security
    Reads user sign-on-name, password and company supplied from the command line or
    from TCSConfig and insert/replace it in both OFS and OFSML request messages types.

.PARAMETER Dummyparm
    Used when the script is called from notepad++ nppexec plugin

.EXAMPLE
    Get-TCSocket.ps1 -Ofstx "OFS.SOURCE,/S/PROCESS,,TCS" -Security
    Description
    -----------
    This command send an OFS transaction request. User info section will be replaced
    with the values defined in TCSConfig script. Response message will be output without
    any formatting to be applied. Results are the same when using '-as raw' option.

.EXAMPLE
    Get-TCSocket.ps1 -Ofstx "ENQUIRY.SELECT,,,CURRENCY-LIST" -Security -Verbose
    Description
    -----------
    This command send an OFS enquiry request. User info section will be replaced
    with the values defined in TCSConfig script. By default, there is no response
    message conversion. -Verbose option outputs additional information showing the
    steps involved in the request processing.

.EXAMPLE
    Get-TCSocket.ps1 -Ofsml "TransInputCustomer.xml" -Security -as raw -pretty
    Description
    -----------
    This command send an OFSML request extracted from specified file name. SecurityContext
    element will be replaced with the values defined in TCSConfig script. Xml response
    message will be "pretty-printed".

.EXAMPLE
    Get-TCSocket.ps1 -Ofsml "StdEnq-AccountList.xml" -Security -as html
    Description
    -----------
    This command send an OFSML request extracted from specified file name. SecurityContext
    element will be replaced with the values defined in TCSConfig script. Xml response
    message will be converted to html format by applying an XSLT stylesheet defined in
    TCSConfig.

.EXAMPLE
    Get-TCSocket.ps1 -Ofstx "ENQUIRY.SELECT,,,INDUSTRY-LIST" -usr T24OFSUSR -pass XXXXXXXX
    Description
    -----------
    This command send an OFS enquiry request. User and password will be used to create
    the T24 session. By default, there is no response message conversion.
    Usr, pass and com are alias for user, password and company named parameters.

.EXAMPLE
    Get-TCSocket.ps1 -Ofstx "ENQUIRY.SELECT,,,SECTOR-LIST" -Security -channel BL.TCS.1
    Description
    -----------
    This command send an OFS enquiry request. User info section will be replaced
    with the values defined in TCSConfig script. By default, there is no response
    message conversion. Connection and timeout settings will be read from 'BL.TCS.1'
    channel name extracted from channels.xml file.

.EXAMPLE
    Get-TCSocket.ps1 -Ofstx "ENQUIRY.SELECT,,,TARGET-LIST" -security -as excel
    Description
    -----------
    This command send an OFS enquiry request. User info section will be replaced
    with the values defined in TCSConfig script. Response message will be converted
    to excel format.

.NOTES
    Name:     Get-TCSocket
    Version:  1.0.0
    Author:   moboan - <moboan> <at> <ambdev> <dot> <net>
    Created:  October 2014
    LastEdit: 2014/10/30

    (C) MOBOAN, Oct 2014 - <moboan> <at> <ambdev> <dot> <net>

    For processing OFS/OFSML request/response messages, Get-TCSocket reads properties
    defined in TCSConfig.ps1 and channels.xml. If a configuration property can
    be provided on both the command line and the config files, then command line
    properties always take precedence.

.LINK
    http://www.youtube.com/user/moboan
.LINK
    https://bitbucket.org/moboan/tcsockets
#>

Param (

    [parameter(
        mandatory = $true,
        ParameterSetName = "Ofstx",
        Position = 0
    )]
    [switch] $Ofstx,

    [parameter(
        mandatory = $true,
        ParameterSetName = "Ofstx",
        Position = 1
    )]
    [ValidateNotNullOrEmpty()]
    [string] $request,

    [parameter(
        mandatory = $true,
        ParameterSetName = "Ofsml",
        Position = 0
    )]
    [switch] $Ofsml,

    [parameter(
        mandatory = $true,
        ParameterSetName = "Ofsml",
        Position = 1
    )]
    [ValidateNotNullOrEmpty()]
    [string] $path,

    [parameter( ParameterSetName = "Ofstx", Position = 2 )]
    [parameter( ParameterSetName = "Ofsml", Position = 2 )]
    [ValidateSet("raw", "plain", "csv", "html", "xml", "excel", "pdf")]
    [string] $as = "raw",

    [ValidateNotNullOrEmpty()]
    [string] $channel,

    [ValidateNotNullOrEmpty()]
    [alias("usr")]
    [string] $user,

    [ValidateNotNullOrEmpty()]
    [alias("pass")]
    [string] $password,

    [ValidateNotNull()]
    [alias("com")]
    [string] $company,

    [switch] $security,

    [parameter( ParameterSetName = "Ofsml" )]
    [switch] $pretty,

    [parameter( ValueFromPipeline = $true )]
    [string] $dummyparm
)

if ($VerbosePreference -eq "Continue") {
    try {
        $max = $host.UI.RawUI.MaxPhysicalWindowSize
        if ( $max ) {
            $host.UI.RawUI.BufferSize = New-Object System.Management.Automation.Host.Size(9999, 9999)
            $host.UI.RawUI.WindowSize = New-Object System.Management.Automation.Host.Size($max.Width, $max.Height)
        }
    } catch {}
}

# load the config file
$scriptDir = Split-Path -Path $MyInvocation.MyCommand.Path -Parent
$ConfigFilePath = Join-Path -Path $scriptDir -ChildPath "Get-TCSConfig.ps1"
$config = (. $ConfigFilePath)

$ModulePath = Join-Path -Path $scriptDir -ChildPath "TCSSession.ps1"
. $ModulePath

switch ($PsCmdlet.ParameterSetName) {

    "Ofstx" { Invoke-OfsRequest   $channel $user $password $company $request $as -security:$security }
    "Ofsml" { Invoke-OfsmlRequest $channel $user $password $company $path    $as -security:$security }
}
