<#
.SYNOPSIS
    Socket session module for Temenos Connector OFS/OFSML requests

.DESCRIPTION
    Contains functions that handle tasks like validating, sending, receiving
    and converting ofs/ofsml requests/responses from/to TC Server.

.NOTES
    Name:     TCSSession
    Version:  1.0.0
    Author:   moboan - <moboan> <at> <ambdev> <dot> <net>
    Created:  October 2014
    LastEdit: 2014/10/30

    (C) MOBOAN, Oct 2014 - <moboan> <at> <ambdev> <dot> <net>
#>


# create log message by combining current datetime and function name
Function New-LogMsg ( $name ) { "$(Get-Date -f $config.DTfmt) [$name] -" }


# open the generated file with the application associated with its extension
# or the application defined in $config.Editor, unless $config.Launch is false
Function Invoke-AssociatedApp {
    Param ( [string] $LiteralPath, [string] $format )
    if ( -not $LiteralPath ) { return }
    if ( -not $config.Launch ) {
        Write-Host "Output file: $([IO.Path]::GetFileName($LiteralPath))"
        return
    }
    Write-Verbose "$(New-LogMsg $MyInvocation.MyCommand.Name) Open '$format' file with the associated program"
    switch -regex ($format) {
        "raw|plain|csv" { if ($config.Editor) { & $config.Editor $LiteralPath } }
                  "xml" { if ($pretty) { ([xml](gc $LiteralPath -ReadCount 0)).Save($LiteralPath) }
                          if ($config.Editor) { & $config.Editor $LiteralPath } }
             "html|pdf" { start $LiteralPath }
                "excel" { [void][Diagnostics.Process]::Start("excel", "$LiteralPath /e") }
    }
}


# build unique path for output file, if -save flag is included in function call
# then save contents to the file.
Function New-OutFilePath {

    Param ( [ref] $contents, [string] $name, [string] $ext,
            [string] $encoding = "UTF-8", [switch] $save )

    if ($contents -and $ext -eq "txt" -and -not $config.GenTxtFile) {
        Write-Verbose "$(New-LogMsg $MyInvocation.MyCommand.Name) Writing output to host..."
        Write-Host $contents.value
        return
    }
    $exportPath = $env:temp
    if ($config.OutPath -and $config.OutPath.Trim()) {
        $exportPath = $config.OutPath
    }
    if (-not (Test-Path $exportPath)) { Write-Host "Invalid output path: $exportPath"; Exit 1 }
    $nameSpec = "$name.$ext"
    if ($config.RandomFn) {
        $rid = [IO.Path]::GetFileNameWithoutExtension([IO.Path]::GetRandomFileName())
        $nameSpec = "${name}_$rid.$ext"
    }
    $fileNameSpec = [IO.Path]::Combine($exportPath, $nameSpec)
    if ($save) {
        $outencoding = [System.Text.Encoding]::GetEncoding($encoding)
        $sw = New-Object System.IO.StreamWriter( $fileNameSpec, $false, $outencoding )
        $sw.Write($contents.value)
        $sw.Close()
    }
    Write-Verbose "$(New-LogMsg $MyInvocation.MyCommand.Name) Output file: $nameSpec"
    $fileNameSpec
}


# send OFS/OFSML requests via sockets to TCServer BytesListener port
Function New-TCRequest {

    Param ( [string] $TCRequest, [string] $TCServer = "localhost",
            [int] $BytesListenerPort = 7099, [int] $Timeout = 50 )

    Write-Verbose "$(New-LogMsg $MyInvocation.MyCommand.Name) Creating new request to ${TCServer}:$BytesListenerPort"
    Write-Verbose "$(New-LogMsg $MyInvocation.MyCommand.Name) Timeout: $Timeout"
    $msgBytes = $Config.Encoding.GetBytes($TCRequest.Trim())

    # OFS/OFSML message format: {message_length}{message}
    # The message length is 4 bytes in the Network bytes order (big-endian)

    # change from little-endian to big-endian byte order - return Int32
    $msgSizeIntBE = [System.Net.IPAddress]::HostToNetworkOrder($msgBytes.Length)
    $msgSizeBytes = [BitConverter]::GetBytes($msgSizeIntBE)

    # build message as required by Temenos Connector BytesListener
    $ofsBytes = new-object System.Byte[] ($msgSizeBytes.Length + $msgBytes.Length)
    [System.Buffer]::BlockCopy($msgSizeBytes, 0, $ofsBytes, 0, $msgSizeBytes.Length)
    [System.Buffer]::BlockCopy($msgBytes, 0, $ofsBytes, $msgSizeBytes.Length, $msgBytes.Length)

    try {
        # create the connection that will be used to send the message
        $socket = New-Object System.Net.Sockets.TcpClient($TCServer, $BytesListenerPort)
        $socket.ReceiveTimeout = $config.ReceiveTimeout
        $stream = $socket.GetStream()
        Write-Verbose "$(New-LogMsg $MyInvocation.MyCommand.Name) TC socket connection is established"
        Start-Sleep -m $Timeout
        $stream.Write($ofsBytes, 0, $ofsBytes.Length)

        $buffSize  = $socket.ReceiveBufferSize
        $buff      = New-Object System.Byte[] $buffSize
        $memStream = New-Object System.IO.MemoryStream($buffSize)
        # read all available data in the response stream
        do {
            Start-Sleep -m $Timeout
            $count = $stream.Read($buff, 0, $buffSize)
            $memStream.Write($buff, 0, $count)
        } while ( $stream.DataAvailable )
        $memStream.Position = 0
        $bytesRead = $memStream.Length - 4

        # read leading length value (4 bytes) and change it from big-endian to little-endian
        # byte order, this give us the expected size of the full response message
        $count         = $memStream.Read($buff, 0, 4)
        $receiveBytes  = New-Object System.IO.MemoryStream($buff, 0, 4)
        $binaryReader  = New-Object System.IO.BinaryReader($receiveBytes)
        $bytesExpected = [Net.IPAddress]::NetworkToHostOrder($binaryReader.ReadInt32())

        $reader = New-Object System.IO.StreamReader($memStream, $Config.Encoding, $false)
        $sb     = New-Object System.Text.StringBuilder($reader.ReadToEnd())

        # check if the incoming message is not truncated
        if (-not ($bytesExpected -eq $bytesRead)) {
            Write-Host "Response message is truncated. Try again or set timeout option to a higher value"
            Write-Host "Receive Len: $bytesExpected; Response Len: $bytesRead"
        }
        Write-Verbose "$(New-LogMsg $MyInvocation.MyCommand.Name) TC socket connection is closed"
        $null = $sb.Replace("`r`n", "").Replace("`n", "")
        $sb.ToString()

    } catch [System.Exception] {
        if ($socket -and $socket.Connected) {
            Write-Host "Connected host has failed to respond, try increasing 'ReceiveTimeout' value in Get-TCSConfig"
        }
        Write-Host $_.Exception.InnerException
        Write-Host $_.InvocationInfo.PositionMessage
        Exit 1
    } finally {
        if ( $stream ) { $stream.Close() }
        if ( $socket ) { $socket.Close() }
        if ( $reader ) { $reader.Close() }
        if ( $memStream ) { $memStream.Close() }
    }
}


# check compliance with OFS syntax requirements
Function Confirm-OfsRequest {

    Param ( [string] $request, [string] $format, [switch] $enquiry )

    Write-Verbose "$(New-LogMsg $MyInvocation.MyCommand.Name) Validating request: $($request.Substring(0, [Math]::Min(60, $request.Length)))..."
    if ($enquiry) {
        # request format: ENQUIRY.SELECT,,USERINFO,IDINFO,DATA
        $m = [regex]::match($request, '^ENQUIRY\.SELECT,,([^\s,]*),([^\s,]+)')
        if (-not $m.Success) {
            Write-Host "the OFS request has missing or invalid parameters"
            return $false
        }
    } else {
        # request format: OPERATION,OPTIONS,USERINFO,IDINFO,DATA
        $m = [regex]::match($request, '^([^\s,]+),([^\s,]*),([^\s,]*),([^\s,]*)')
        if (-not $m.Success) {
            Write-Host "the OFS request has missing or invalid parameters"
            return $false
        }
    }
    if (($format -eq "xml" ) -or ($format -eq "html")) {
        Write-Host "'$format' format not implemented for this request type"
        return $false
    }
    Write-Verbose "$(New-LogMsg $MyInvocation.MyCommand.Name) Is request valid: $true"
    $true
}


# check if the response from TCServer contains valid data
Function Confirm-OfsResponse {

    Param ( [ref] $response, [switch] $enquiry )

    Write-Verbose "$(New-LogMsg $MyInvocation.MyCommand.Name) Validating response..."

    if (-not $enquiry) {
        $m = [regex]::match($response.value, '^([^\s/]*)/([^\s/]*)/([1-3]{1})')
        if (-not $m.Success) {
            Write-Verbose "$(New-LogMsg $MyInvocation.MyCommand.Name) Response has missing or invalid data"
            Write-Host $response.value
        }
        return $m.Success
    }
    # check response to see if enquiry request contents were valid
    $idx = $response.value.IndexOf(',"')
    if (-not ($idx -gt 0)) {
        Write-Verbose "$(New-LogMsg $MyInvocation.MyCommand.Name) Response has missing or invalid data"
        Write-Host $response.value
        return $false
    }
    $hdrSection = $response.value.Substring(0, $idx)
    $hdrRaw = $hdrSection.split(",", 3)
    if (-not ($hdrRaw.count -gt 1 -and $hdrRaw.count -lt 4)) {
        Write-Verbose "$(New-LogMsg $MyInvocation.MyCommand.Name) Header section has missing or invalid data"
        Write-Host $response.value
        return $false
    }
    # check if the enquiry response is empty - no records
    $offset = $idx + 1
    $dataSection = $response.value.Substring($offset, [Math]::Min(30, $response.value.Length - $offset))
    if ($dataSection.StartsWith('"No records were found')) {
        Write-Verbose "$(New-LogMsg $MyInvocation.MyCommand.Name) No records were found"
        $dataSection = $response.value.Substring($offset)
        Write-Host ( $dataSection.split(",") | % { $_.Trim('"').Trim() } )
        return $false
    }

    Write-Verbose "$(New-LogMsg $MyInvocation.MyCommand.Name) Is response valid: $true"
    $true
}


# transformation of ENQUIRY.SELECT response to plain text file or csv file
# response format: HEADER CAPTION DETAILS,COLUMN DETAILS,RESPONSE DATA
Function Convert-OfsResponse {

    Param ( [ref] $response, [string] $title = "OfsEnqResponse", [switch] $csv )

    Write-Verbose "$(New-LogMsg $MyInvocation.MyCommand.Name) Transforming response..."

    # split the response between header section and data section
    $idx = $response.value.IndexOf(',"')
    $hdrSection = $response.value.Substring(0, $idx)
    $dataSection = $response.value.Substring($idx + 1)

    # build header and column headings
    $hdrRaw = $hdrSection.split(",", 3)

    # header caption details format: Identifier1=Text1/Identifier2=Text2
    $captionDetails = $hdrRaw[$hdrRaw.count - 2].Trim().StartsWith("HEADER=")
    if ($captionDetails) {
        Write-Warning "Enquiries that have caption details might display incorrect output"
        Write-Warning "Use the 'raw' format option to see the original response message"
        $svm = [string][char]0xF8FC
        $captions = $hdrRaw[$hdrRaw.count - 2].Trim() -split '/+[^=\s]+='
        for ( $i = 0; $i -lt $captions.Count; $i++ ) {
            $mark = $captions[$i].IndexOf('=')
            $name = $captions[$i].Substring($mark + 1).Trim()
            if ($name -match $svm) { $name = $name.Replace($svm, " ") }
            if ($name -match '^".*"$') {
                $prefix = ","
                if ($i -eq 0 -and $dataSection.StartsWith($name)) { $prefix = "" }
                $name = '"' + $name.Trim('"') + '\s*"'
                if ([regex]::match($dataSection, $name).success) {
                    $name = $prefix + $name
                } else { $name = ""}
            } else {
                $name = ""
            }
            $captions[$i] = $name
        }
        $matchCaption = $false
        $patCaption = ($captions -join "")
        if ($patCaption.Trim()) {
            $patCaption += ","
            $replacement = ""
            if ($patCaption.StartsWith(",")) { $replacement = "`t" }
            $matchCaption = [regex]::match($dataSection, $patCaption).success
            if (-not $matchCaption) {
                $patCaption = $patCaption -replace '","', """`t"""
                $matchCaption = [regex]::match($dataSection, $patCaption).success
            }
            if ($matchCaption) { $dataSection = $dataSection -replace $patCaption, $replacement }
        }
        Write-Verbose "$(New-LogMsg $MyInvocation.MyCommand.Name) Caption pattern: '$($patCaption)'"
        Write-Verbose "$(New-LogMsg $MyInvocation.MyCommand.Name) Caption match: $($matchCaption)"
    }

    # column details format: identifier:format type:label/identifier:format type:label
    # eg., REFERENCE.CCY:ALPHANUMERIC:Reference Ccy/VALUATION.AMT:AMOUNT:VALUATION.AMT
    $columnDetails =  $hdrRaw[$hdrRaw.count - 1].Trim()
    $headings = $columnDetails.Split("/")
    for ( $i = 0; $i -lt $headings.Count; $i++ ) {
        $name = $headings[$i].split(":")
        # if label is missing use identifier instead
        if ($name -and $name[2] -and $name[2].Trim()) { $pos = 2 } else { $pos = 0 }
        $headings[$i] = $name[$pos].Trim()
    }

    # transform the data section of the response
    if ([regex]::Match($dataSection, '\t').Success) {
        $patternOld = '(?<=\t")([^"\t]*?"),"'
    } else {
        $patternOld = '(?<=")([^"\t]*?"),"'
    }
    $patternNew = "`$1`t`r`n"""
    $dataSection = $dataSection -replace $patternOld, $patternNew
    $contents = $dataSection -split "`r`n"
    $contents[$contents.Count - 1] += "`t"

    # get occurrences of the number of columns in the response data section
    # enquiries with HEADER, FOOTER, BREAK or TOTAL attributes will have rows
    # with different number of columns
    $colCountOccurr = @()
    for ( $i = 0; $i -lt $contents.Count; $i++ ) {
        $nCols = $contents[$i].Split("`t").Count - 1
        if (-not ($colCountOccurr -match $nCols)) { $colCountOccurr += $nCols }
    }
    Write-Verbose "$(New-LogMsg $MyInvocation.MyCommand.Name) Column count occurrences: $($colCountOccurr.Count)"

    if ($colCountOccurr.Count -eq 1) {
        # get the number of columns in the response data section
        $nCols = $contents[0].Split("`t").Count - 1

        # if empty column headers are found, dynamically fill those missing
        [System.Collections.ArrayList] $headingsNew = @()
        for ( $i = 0; $i -lt $nCols; $i++ ) {
            if ([string]::IsNullOrEmpty($headings[$i])) {
                $colName = "COLUMN $($i + 1)"
            } else {
                $colName = $headings[$i]
            }
            $null = $headingsNew.Add($colName)
        }
        $headings = $headingsNew.ToArray()
    }

    # build and export the output in the requested format
    $sb = New-Object System.Text.StringBuilder
    $headings | % { $null = $sb.Append($_).Append("`t") }
    $null = $sb.AppendLine()

    # format and save as plain text
    if (-not $csv) {
        $null = $sb.Append($dataSection).Append("`t")
        Write-Verbose "$(New-LogMsg $MyInvocation.MyCommand.Name) Response transformed"
        return New-OutFilePath ([ref]$sb.ToString()) $title "txt" -save
    }

    # format and save as csv text using tab as delimiter
    foreach ($row in $contents) {
        $columns = $row -split '"\t'
        for ( $i = 0; $i -lt $columns.count - 1; $i++ ) {
            $null = $sb.Append( $columns[$i].Substring(1).Trim() ).Append("`t")
        }
        $null = $sb.AppendLine()
    }
    Write-Verbose "$(New-LogMsg $MyInvocation.MyCommand.Name) Response transformed"
    New-OutFilePath ([ref]$sb.Replace('|', '').ToString()) $title "csv" "UNICODE" -save
}



# transform ofsml response to html format using XSLT stylesheet
Function Convert-OfsmlResponse {
    Param ( [ref] $response )

    Write-Verbose "$(New-LogMsg $MyInvocation.MyCommand.Name) Transforming response..."
    Try {
        $xresp = new-object System.Xml.XmlDocument
        $xresp.loadXml($response.value)
        $ofsFaultCode = $xresp.T24.serviceResponse.ofsFault.ofsFaultCode
        $ofsFaultMessage = $xresp.T24.serviceResponse.ofsFault.ofsFaultMessage
        if ($ofsFaultCode -or $ofsFaultMessage) {
            Write-Host "ofsFaultCode: $ofsFaultCode`r`nofsFaultMessage: $ofsFaultMessage"
            return $null
        }
        $enq = $xresp.T24.serviceResponse.ofsStandardEnquiry
        if (-not ($enq.status -eq "OK")) {
            Write-Host "Response has missing or invalid data: $($enq.status)"
            return $null
        }
        Write-Verbose "$(New-LogMsg $MyInvocation.MyCommand.Name) Enquiry name: $($enq.name) status: $($enq.status)"
        $xmlBytes = $Config.Encoding.GetBytes($response.value)
        $memStream = New-Object System.IO.MemoryStream(,$xmlBytes)
        $sr = New-Object System.IO.StreamReader($memStream)
        $xpathDoc = New-Object System.Xml.XPath.XPathDocument($sr)
        $xslt = New-Object System.Xml.Xsl.XslCompiledTransform
        $xslt.Load($config.Xsl)
        $invalidChars = "[{0}]" -f ([regex]::Escape([IO.Path]::GetInvalidFileNameChars() -join ''))
        $validName = [regex]::Replace($enq.name, $invalidChars, '-')
        $htmlFile = New-OutFilePath $null $validName "html"
        $sw = New-Object System.IO.StreamWriter $htmlFile
        $xslt.Transform($xpathDoc, $null, $sw)
        $htmlFile

    } catch [System.Exception] {
        Write-Host $error[0]
        Write-Host $_.InvocationInfo.PositionMessage
        Exit 1
    } finally {
        if ($sr) { $sr.close() }
        if ($sw) { $sw.close() }
    }
}


# replace "securityContext" node in ofsml request with UserInfo
# provided in command line or defined in Get-TCSConfig
Function Set-OfsmlSecurityContext {

    Param ( [System.Xml.XmlDocument] $xdoc,  [Hashtable] $UserInfo)

    Write-Verbose "$(New-LogMsg $MyInvocation.MyCommand.Name) Replacing 'securityContext'..."

    $nsURI = $xdoc.DocumentElement.NamespaceURI
    $nsManager = New-Object System.Xml.XmlNamespaceManager($xdoc.NameTable)
    $nsManager.AddNamespace("ns", $nsURI)

    $newSecCtx = $xdoc.CreateElement("securityContext", $nsURI)
    $newSecCtx.SetAttribute("encrypt", "false")
    $user = $xdoc.CreateElement("userName", $nsURI)
    if ($UserInfo.User) { $user.InnerText = $UserInfo.User }
    $pass  = $xdoc.CreateElement("password", $nsURI)
    if ($UserInfo.Pass) { $pass.InnerText = $UserInfo.Pass }
    $cpny = $xdoc.CreateElement("company", $nsURI)
    if ($UserInfo.Company) { $cpny.InnerText = $UserInfo.Company }
    $null = $newSecCtx.AppendChild($user)
    $null = $newSecCtx.AppendChild($pass)
    $null = $newSecCtx.AppendChild($cpny)

    # insert namespace in xpath selection and replace securitycontext
    $xpathSel  = "/T24/serviceRequest" -replace "/", "/ns:"
    $servReq   = $xdoc.SelectSingleNode($xpathSel, $nsManager)
    $xpathSel  = "/T24/serviceRequest/securityContext" -replace "/", "/ns:"
    $oldSecCtx = $xdoc.SelectSingleNode($xpathSel, $nsManager)
    $null      = $servReq.ReplaceChild($newSecCtx, $oldSecCtx)

    Write-Verbose "$(New-LogMsg $MyInvocation.MyCommand.Name) 'securityContext' replaced"
    $xdoc
}


# create an excel file using the excel application and tools
# transform input csv file to excel applying a predefined table style format
# a few available customizations are defined in Get-TCSConfig
Function ConvertTo-Excel {

    Param ( [string] $csvfile )

    $xlWorkbookDefault = 51
    $xlYes = 1
    try {
        Write-Verbose "$(New-LogMsg $MyInvocation.MyCommand.Name) Creating excel file..."

        $excel = New-Object Microsoft.Office.Interop.Excel.ApplicationClass
        $excel.visible = $false
        $excel.DisplayAlerts = $false

        $Workbook = $excel.Workbooks.Open($csvfile)
        $Worksheet = $Workbook.Worksheets.Item(1)
        $UsedRange = $Worksheet.UsedRange

        $ListObject = $Worksheet.ListObjects.Add(1, $UsedRange, $null, $xlYes)
        $ListObject.TableStyle = $config.xlTableStyle
        # convert table to data range
        if ($config.xlDataRange) { $ListObject.Unlist() }

        $excel.ActiveWindow.Zoom = $Config.xlZoom
        $null = $UsedRange.Columns.AutoFit()
        $range = $Worksheet.range("A2")
        $null = $range.Select()
        if ($config.xlDataRange) { $excel.ActiveWindow.FreezePanes = $true }

        $xlFileName = [IO.Path]::ChangeExtension($csvfile, "xlsx")
        $Workbook.SaveAs($xlFileName, $xlWorkbookDefault)

        Write-Verbose "$(New-LogMsg $MyInvocation.MyCommand.Name) Output file: $([IO.Path]::GetFileName($xlFileName))"
        $xlFileName

    } catch [System.Exception] {
        Write-Host $error[0]
        Write-Host $_.InvocationInfo.PositionMessage
        return $null
    } Finally {
        if ($excel) {
            $excel.Workbooks.close()
            $excel.Quit()
            [void][System.Runtime.Interopservices.Marshal]::ReleaseComObject($excel)
            Remove-Variable -Name excel
        }
    }
}



# create a pdf file using the excel application and tools
# transform input csv file to excel applying a predefined table style format
# the resulting excel file is then saved as pdf format
Function ConvertTo-Pdf {

    Param ( [string] $csvfile )

    $xlPaperLetter = 1
    $xlPortrait = 1
    $xlQualityStandard = 0
    $xlTypePDF = 0
    $xlWorkbookDefault = 51
    $xlYes = 1
    try {
        Write-Verbose "$(New-LogMsg $MyInvocation.MyCommand.Name) Creating excel file..."

        $excel = New-Object Microsoft.Office.Interop.Excel.ApplicationClass
        $excel.visible = $false
        $excel.DisplayAlerts = $false

        $Workbook = $excel.Workbooks.Open($csvfile)
        $Worksheet = $Workbook.Worksheets.Item(1)
        $UsedRange = $Worksheet.UsedRange

        $ListObject = $Worksheet.ListObjects.Add(1, $UsedRange, $null, $xlYes)
        $ListObject.TableStyle = $config.xlTableStyle
        $null = $UsedRange.Columns.AutoFit()

        $Worksheet.PageSetup.PrintArea = $UsedRange.Address()
        $Worksheet.PageSetup.PrintTitleRows = '$1:$1'
        $Worksheet.PageSetup.CenterHeader = $Worksheet.Name
        $Worksheet.PageSetup.CenterFooter = "Page &P of &N"
        $Worksheet.PageSetup.LeftMargin = $excel.InchesToPoints(0.25)
        $Worksheet.PageSetup.RightMargin = $excel.InchesToPoints(0.25)
        $Worksheet.PageSetup.TopMargin = $excel.InchesToPoints(0.5)
        $Worksheet.PageSetup.BottomMargin = $excel.InchesToPoints(0.5)
        $Worksheet.PageSetup.HeaderMargin = $excel.InchesToPoints(0.25)
        $Worksheet.PageSetup.FooterMargin = $excel.InchesToPoints(0.25)
        $Worksheet.PageSetup.PrintGridlines = $false
        $Worksheet.PageSetup.CenterHorizontally = $true
        $Worksheet.PageSetup.Orientation = $xlPortrait
        $Worksheet.PageSetup.PaperSize = $xlPaperLetter

        $Worksheet.PageSetup.Zoom = $false
        $Worksheet.PageSetup.FitToPagesWide = 1
        $Worksheet.PageSetup.FitToPagesTall = 9999
        $Worksheet.PageSetup.ScaleWithDocHeaderFooter = $true
        $Worksheet.PageSetup.AlignMarginsHeaderFooter = $false

        $xlFileName = [IO.Path]::ChangeExtension($csvfile, "xlsx")
        $Workbook.SaveAs($xlFileName, $xlWorkbookDefault)
        $pdfFileName = [IO.Path]::ChangeExtension($xlFileName, "pdf")
        Write-Verbose "$(New-LogMsg $MyInvocation.MyCommand.Name) Creating pdf file..."
        $workbook.ExportAsFixedFormat($xlTypePDF, $pdfFileName, $xlQualityStandard)

        Write-Verbose "$(New-LogMsg $MyInvocation.MyCommand.Name) Output file: $([IO.Path]::GetFileName($pdfFileName))"
        # returns pdf and excel filenames in array
        $pdfFileName, $xlFileName

    } catch [System.Exception] {
        Write-Host $error[0]
        Write-Host $_.InvocationInfo.PositionMessage
        return $null
    } Finally {
        if ($excel) {
            $excel.Workbooks.close()
            $excel.Quit()
            [void][System.Runtime.Interopservices.Marshal]::ReleaseComObject($excel)
            Remove-Variable -Name excel
        }
    }
}


# get channel info from channels.xml file and populate hashtable with
# TCServer connection settings
Function Read-TCChannel ( [string] $channelID , [string] $LiteralPath ) {
    Write-Verbose "$(New-LogMsg $MyInvocation.MyCommand.Name) Reading TC settings from channel: $channelID"
    try {
        $xdoc = new-object System.Xml.XmlDocument
        $xdoc.load($LiteralPath)
        $channels = $xdoc.SelectNodes("//CHANNEL")
        $channels.GetEnumerator() | ? {$_.Name.Trim() -eq $channelID} | % { $node = $_ }
        if ($node) {
            $channel = @{
                "Host"    = $node.ADAPTER.SUPPLIER.INITIATOR.HOSTNAME.Trim()
                "Port"    = [int] $node.ADAPTER.PORT
                "Timeout" = [int] $node.TIMEOUT
            }
            $channel
        } else {
            throw "Channel '$channelID' not found"
        }
    } catch [System.Exception] {
        Write-Host $error[0]
        Write-Host $_.InvocationInfo.PositionMessage
        Exit 1
    }
}


# Process OFS requests
Function Invoke-OfsRequest {

    Param ( [string] $channel, [string] $user, [string] $pass, [string] $company,
            [string] $request, [string] $format, [switch] $security )

    Write-Verbose "$(New-LogMsg $MyInvocation.MyCommand.Name) Request processing start"
    $enquiry = $request.StartsWith("ENQUIRY.SELECT,,", "CurrentCultureIgnoreCase")
    if ( -not (Confirm-OfsRequest $request $format -enquiry:$enquiry) ) { return }
    if (-not $enquiry -and $format -ne "raw") {
        Write-Verbose "$(New-LogMsg $MyInvocation.MyCommand.Name) '$format' format not implemented, changed to 'raw'"
        $format = "raw"
    }
    if ($security) {
        # replace/insert user info into the OFS request
        $userInfo = "{0}/{1}/{2}" -f $user, $pass, $company
        $request = (,($request -split ",") | % { $_[2] = $userInfo; ,($_) }) -join ","
    }

    $node = Read-TCChannel $channel $config.ChannelFile
    $response = New-TCRequest $request $node.Host $node.Port $node.Timeout
    if ( -not ( Confirm-OfsResponse ([ref]$response) -enquiry:$enquiry ) ) { return }
    if ($format -eq "raw") {
        $rawFile = New-OutFilePath ([ref]$response) "OfsRawResp" "txt" -save
        if ($rawFile) { Invoke-AssociatedApp $rawFile $format }
        return
    }
    # from this point forward only 'ENQUIRY.SELECT' type responses
    # will be transformed into the specified output format

    # extract Enquiry ID from ofs request
    $title = $request.split(",", 5)[3].Trim()

    # checking file name for invalid characters
    $invalidChars = "[{0}]" -f ([regex]::Escape([IO.Path]::GetInvalidFileNameChars() -join ''))
    $title = [regex]::Replace($title, $invalidChars, '-')

    if ($format -eq "plain") {
        $plainFile = Convert-OfsResponse ([ref]$response) $title
        Invoke-AssociatedApp $plainFile $format
        return
    }
    $csvFile = Convert-OfsResponse ([ref]$response) $title -csv
    if ($format -eq "csv") {
        Invoke-AssociatedApp $csvFile $format
        return
    }

    # load the primary interop assemblies (PIAs) for Excel
    Add-Type -Assembly "Microsoft.Office.Interop.Excel"

    if ($format -eq "excel") {
        $xlFile  = ConvertTo-Excel $csvFile
        if (-not $xlFile) { return }
        if (-not $config.Keep) { Remove-Item $csvFile -EA SilentlyContinue }
        Invoke-AssociatedApp $xlFile $format
    } elseif ($format -eq "pdf") {
        $xlPdfFile = ConvertTo-Pdf $csvFile
        if (-not $xlPdfFile) { return }
        if (-not $config.Keep) { Remove-Item $csvFile, $xlPdfFile[1] -EA SilentlyContinue }
        Invoke-AssociatedApp $xlPdfFile[0] $format
    }
}


# Process OFSML requests
Function Invoke-OfsmlRequest {

    Param ( [string] $channel, [string] $user, [string] $pass, [string] $company,
            [string] $LiteralPath, [string] $format, [switch] $security )

    if ($LiteralPath -and $LiteralPath.Trim().Length -eq 0) { return }
    Write-Verbose "$(New-LogMsg $MyInvocation.MyCommand.Name) Request processing start"
    if (-not ($format -eq "raw" -or $format -eq "html")) {
        Write-Host "'$format' format not implemented for this request type"
        return
    }
    try {
        $xmlFileInfo = Get-item $LiteralPath -ErrorAction Stop
        $xdoc = new-object System.Xml.XmlDocument
        $xdoc.load($xmlFileInfo.FullName)
        $stdEnq = $xdoc.T24.serviceRequest.ofsStandardEnquiry
        if ($security) {
            # replace/insert user info into the OFSML request
            $userInfo = @{ "User" = $user; "Pass" = $pass; "Company" = $company }
            $xdoc = Set-OfsmlSecurityContext $xdoc $userInfo
        }
        $request = $xdoc.InnerXml
    } catch [System.Exception] {
        Write-Host $error[0]
        Write-Host $_.InvocationInfo.PositionMessage
        Exit 1
    }

     if (-not $stdEnq -and $format -eq "html") {
        Write-Host "'$format' format not implemented for this request type"
        return
    }

    $node = Read-TCChannel $channel $config.ChannelFile
    $response = New-TCRequest $request $node.Host $node.Port $node.Timeout
    if ($format -eq "raw") {
        $xmlFile = New-OutFilePath ([ref]$response) "OfsmlResp" "xml" -save
        if (-not $xmlFile) { return }
        Invoke-AssociatedApp $xmlFile "xml"
    } elseif ($format -eq "html") {
        $htmlFile = Convert-OfsmlResponse ([ref]$response)
        if (-not $htmlFile) { return }
        Invoke-AssociatedApp $htmlFile $format
    }
}
